@@ -17,6 +17,7 @@ import argparse
 import contextlib
 import datetime
 import itertools
+import json
 import logging
 import operator
 import shutil
@@ -57,7 +58,22 @@ def _prod(xs: t.Iterable[int]) -> int:
     return reduce(operator.mul, xs, 1)
 
 
-def __open_dataset_file(filename: str) -> xr.Dataset:
+def _make_grib_dataset_inmem(grib_ds: xr.Dataset) -> xr.Dataset:
+    # Copies all the vars to in-memory to reduce disk seeks everytime a single row is processed.
+    # This also removes the need to keep the backing temp source file around.
+    data_ds = grib_ds.copy(deep=True)
+    for v in grib_ds.variables:
+        if v not in data_ds.coords:
+            data_ds[v].variable.values = grib_ds[v].variable.values
+    return data_ds
+
+
+def __open_dataset_file(filename: str, open_dataset_kwargs: t.Optional[t.Dict] = None) -> xr.Dataset:
+
+    if open_dataset_kwargs:
+        return _make_grib_dataset_inmem(xr.open_dataset(filename, **open_dataset_kwargs))
+
+    # If no open kwargs are available, make educated guesses about the dataset.
     try:
         return xr.open_dataset(filename)
     except ValueError as e:
@@ -66,7 +82,8 @@ def __open_dataset_file(filename: str) -> xr.Dataset:
             raise
     # Trying with explicit engine for cfgrib.
     try:
-        return xr.open_dataset(filename, engine='cfgrib', backend_kwargs={'indexpath': ''})
+        return _make_grib_dataset_inmem(
+            xr.open_dataset(filename, engine='cfgrib', backend_kwargs={'indexpath': ''}))
     except ValueError as e:
         if "multiple values for key 'edition'" not in str(e):
             raise
@@ -75,12 +92,12 @@ def __open_dataset_file(filename: str) -> xr.Dataset:
     # Try with edition 1
     # Note: picking edition 1 for now as it seems to get the most data/variables for ECMWF realtime data.
     # TODO(#54): Make this a more generic function that can take custom args from tool users.
-    return xr.open_dataset(filename, engine='cfgrib',
-                           backend_kwargs={'filter_by_keys': {'edition': 1}, 'indexpath': ''})
+    return _make_grib_dataset_inmem(xr.open_dataset(filename, engine='cfgrib',
+                                                    backend_kwargs={'filter_by_keys': {'edition': 1}, 'indexpath': ''}))
 
 
 @contextlib.contextmanager
-def open_dataset(uri: str) -> t.Iterator[xr.Dataset]:
+def open_dataset(uri: str, open_dataset_kwargs: t.Optional[t.Dict] = None) -> t.Iterator[xr.Dataset]:
     """Open the dataset at 'uri' and return a xarray.Dataset."""
     try:
         # Copy netcdf or grib object from cloud storage, like GCS, to local file
@@ -91,7 +108,7 @@ def open_dataset(uri: str) -> t.Iterator[xr.Dataset]:
                 shutil.copyfileobj(source_file, dest_file)
                 dest_file.flush()
                 dest_file.seek(0)
-                xr_dataset: xr.Dataset = __open_dataset_file(dest_file.name)
+                xr_dataset: xr.Dataset = __open_dataset_file(dest_file.name, open_dataset_kwargs)
 
                 logger.info(f'opened dataset size: {xr_dataset.nbytes}')
 
@@ -142,8 +159,9 @@ def to_table_schema(columns: t.List[t.Tuple[str, str]]) -> t.List[bigquery.Schem
 
 def to_json_serializable_type(value: t.Any) -> t.Any:
     """Returns the value with a type serializable to JSON"""
+    # Note: The order of processing is significant.
     logger.debug('Serializing to JSON')
-    if type(value) == np.float32 or type(value) == np.float64:
+    if np.issubdtype(type(value), np.floating):
         return float(value)
     elif type(value) == np.ndarray:
         # Will return a scaler if array is of size 1, else will return a list.
@@ -173,7 +191,11 @@ def to_json_serializable_type(value: t.Any) -> t.Any:
         return None
     elif type(value) == np.timedelta64:
         # Return time delta in seconds.
-        return value / np.timedelta64(1, 's')
+        return float(value / np.timedelta64(1, 's'))
+    # This check must happen after processing np.timedelta64 and np.datetime64.
+    elif np.issubdtype(type(value), np.integer):
+        return int(value)
+
     return value
 
 
@@ -185,7 +207,8 @@ def _only_target_vars(ds: xr.Dataset, data_vars: t.Optional[t.List[str]] = None)
         logger.info(f'target data_vars empty; using whole dataset; size: {ds.nbytes}')
         return ds
 
-    assert all([dv in ds.data_vars for dv in data_vars]), 'Target variable must be in original dataset.'
+    assert all([dv in ds.data_vars or dv in ds.coords for dv in data_vars]), 'Target variable must be in original ' \
+                                                                             'dataset. '
 
     dropped_ds = ds.drop_vars([v for v in ds.data_vars if v not in data_vars])
     logger.info(f'target-only dataset size: {dropped_ds.nbytes}')
@@ -193,7 +216,7 @@ def _only_target_vars(ds: xr.Dataset, data_vars: t.Optional[t.List[str]] = None)
     return dropped_ds
 
 
-def get_coordinates(ds: xr.Dataset) -> t.Iterator[t.Dict]:
+def get_coordinates(ds: xr.Dataset, uri: str = '') -> t.Iterator[t.Dict]:
     """Generates normalized coordinate dictionaries that can be used to index Datasets with `.loc[]`."""
     # Creates flattened iterator of all coordinate positions in the Dataset.
     #
@@ -218,16 +241,17 @@ def get_coordinates(ds: xr.Dataset) -> t.Iterator[t.Dict]:
     total_coords = _prod(ds.coords.dims.values())
     for idx, it in enumerate(coords):
         if idx % 1000 == 0:
-            logger.info(f'Processed {idx // 1000}k coordinates of {(total_coords / 1000):2f}k...')
+            logger.info(f'Processed {idx // 1000}k / {(total_coords / 1000):.2f}k coordinates for {uri!r}...')
         yield dict(zip(ds.coords.indexes, it))
 
-    logger.info(f'Finished processing all {(idx / 1000):2f}k coordinates.')
+    logger.info(f'Finished processing all {(idx / 1000):.2f}k coordinates.')
 
 
 def extract_rows(uri: str, *,
                  variables: t.Optional[t.List[str]] = None,
                  area: t.Optional[t.List[int]] = None,
-                 import_time: t.Optional[str] = DEFAULT_IMPORT_TIME) -> t.Iterator[t.Dict]:
+                 import_time: t.Optional[str] = DEFAULT_IMPORT_TIME,
+                 open_dataset_kwargs: t.Optional[t.Dict] = None) -> t.Iterator[t.Dict]:
     """Reads named netcdf then yields each of its rows as a dict mapping column names to values."""
     logger.info(f'Extracting rows as dicts: {uri!r}.')
 
@@ -235,7 +259,7 @@ def extract_rows(uri: str, *,
     if not import_time:
         import_time = datetime.datetime.utcnow().replace(tzinfo=datetime.timezone.utc).isoformat()
 
-    with open_dataset(uri) as ds:
+    with open_dataset(uri, open_dataset_kwargs) as ds:
         data_ds: xr.Dataset = _only_target_vars(ds, variables)
 
         if area:
@@ -262,8 +286,9 @@ def extract_rows(uri: str, *,
             row.update(it)
             # Add un-indexed coordinates.
             for c in row_ds.coords:
-                if c not in it:
+                if c not in it and (not variables or c in variables):
                     row[c] = to_json_serializable_type(ensure_us_time_resolution(row_ds[c].values))
+
             # Add import metadata.
             row[DATA_IMPORT_TIME_COLUMN] = import_time
             row[DATA_URI_COLUMN] = uri
@@ -275,7 +300,7 @@ def extract_rows(uri: str, *,
             beam.metrics.Metrics.counter('Success', 'ExtractRows').inc()
             return row
 
-        yield from map(to_row, get_coordinates(data_ds))
+        yield from map(to_row, get_coordinates(data_ds, uri))
 
 
 def pattern_to_uris(match_pattern: str) -> t.Iterable[str]:
@@ -295,8 +320,8 @@ def run(argv: t.List[str], save_main_session: bool = True):
                         help="Full name of destination BigQuery table (<project>.<dataset>.<table>). Table will be "
                              "created if it doesn't exist.")
     parser.add_argument('-v', '--variables', metavar='variables', type=str, nargs='+', default=list(),
-                        help='Target variables for the BigQuery schema. Default: will import all data variables as '
-                             'columns.')
+                        help='Target variables (or coordinates) for the BigQuery schema. Default: will import all '
+                             'data variables as columns.')
     parser.add_argument('-a', '--area', metavar='area', type=int, nargs='+', default=list(),
                         help='Target area in [N, W, S, E]. Default: Will include all available area.')
     parser.add_argument('--topic', type=str,
@@ -313,6 +338,8 @@ def run(argv: t.List[str], save_main_session: bool = True):
                               "time (format: YYYY-MM-DD HH:MM:SS.usec+offset). Default: now in UTC."))
     parser.add_argument('--infer_schema', action='store_true', default=False,
                         help='Download one file in the URI pattern and infer a schema from that file. Default: off')
+    parser.add_argument('--xarray_open_dataset_kwargs', type=json.loads, default='{}',
+                        help='Keyword-args to pass into `xarray.open_dataset()` in the form of a JSON string.')
     parser.add_argument('-d', '--dry-run', action='store_true', default=False,
                         help='Preview the load into BigQuery. Default: off')
 
@@ -348,7 +375,7 @@ def run(argv: t.List[str], save_main_session: bool = True):
         )
     else:
         logger.info('Inferring schema from data.')
-        with open_dataset(next(iter(all_uris))) as open_ds:
+        with open_dataset(next(iter(all_uris)), known_args.xarray_open_dataset_kwargs) as open_ds:
             ds: xr.Dataset = _only_target_vars(open_ds, known_args.variables)
             table_schema = dataset_to_table_schema(ds)
 
@@ -385,12 +412,12 @@ def run(argv: t.List[str], save_main_session: bool = True):
                     extract_rows,
                     variables=known_args.variables,
                     area=known_args.area,
-                    import_time=known_args.import_time)
+                    import_time=known_args.import_time,
+                    open_dataset_kwargs=known_args.xarray_open_dataset_kwargs)
                 | 'WriteToBigQuery' >> WriteToBigQuery(
                     project=table.project,
                     dataset=table.dataset_id,
                     table=table.table_id,
-                    method='STREAMING_INSERTS',
                     write_disposition=BigQueryDisposition.WRITE_APPEND,
                     create_disposition=BigQueryDisposition.CREATE_NEVER)
         )
